# explorative modelling: predefined operating scheme 2
import pandas as pd
import numpy as np
import numpy.random
from numpy import arange
import numpy_financial as npf
from numpy_financial import npv, irr
import xlwt
from openpyxl import load_workbook
import os
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
import sys
import math

# locate file and load the workbook
directory = "/home/jessieluo" # for example
os.chdir(directory)
file_name = "process results.xlsm"
file = directory + "/" + file_name
wb1 = load_workbook(file_name, keep_vba=True, data_only=True)
wb = load_workbook(file, keep_vba=True)

# plant information
min_operatingload = 0.70  # based on the volume flexibility results at the equipment level
ws_elec = wb1['Overall electricity balance']
MLload_100 = ws_elec['A27'].value  # nominal electricity consumption, GJ/h
ws_economics = wb1["CAPEX & OPEX"]
ws_ecoassumption = wb1['Economic assumptions'] # a hidden spreadsheet
lifetime = ws_ecoassumption['B7'].value  # life time, y
discount_rate = ws_ecoassumption['B9'].value
salvage_rate = ws_ecoassumption['B10'].value
projection_2030 = 1  # from NREL
storage_info = {"type": ["Li", "Lead", "RFB"],
                "unit_cost": [1200/3.6/1000*projection_2030, 500/3.6/1000*projection_2030, 1000/3.6/1000*projection_2030],
                "efficiency": [0.9, 0.85, 0.85],
                "salvage_rate": [0.05, 0.05, 0.05]} # unit cost: M€/GJ

# read CAPEX and OPEX parameters
# It and OMt are fixed at the 100% load
It_ML_ref = [0 for i in range(lifetime+1)] # CAPEX flow, M€/y
It_ML_ref[0] = ws_economics['G18'].value  # specify the initial CAPEX, M€/y
It_ML_ref[15] = (ws_ecoassumption['M59'].value + ws_ecoassumption['N59'].value) / 1e6  # specify the replacement of the membrane contactor
It_ML_ref[30] = - It_ML_ref[0] * salvage_rate  # specify the salvage value M€
OMt_ML_load100 = ws_economics['G22'].value # operation and maintenance cost at 100% load, M€/y

# Wt, Ft, Et are always linear to production rate; no need to be adjusted based on power profile
ws_mass = wb1['Overall mass balance']
Wt_ML_load100 = ws_mass['B57'].value  # waste treatment cost at 100% load, €/h
Ft_ML_load100 = ws_mass['B56'].value  # feedstock cost at 100% load, €/h
Et_ML_load100 = ws_elec['A31'].value  # electricity cost at 100% load, €/h
Mt_ML_load100 = ws_mass['B65'].value  # production of C6A at 100% load, kg/h

# Heat consumption is nonlinear to production rate, €/h
ws_heat = wb1["Overall utility balance"]
LR_list = []
for row in ws_heat['B72':'AF72']:
    for x in row:
        LR_list.append(x.value)
Ht_cost_list1 = []
for row in ws_heat['B81':'AF81']:
    for x in row:
        Ht_cost_list.append(round(x.value, 2))
Ht_list = {0: 0}
for i in range(len(LR_list)):
    Ht_list.update({LR_list[i]: Ht_cost_list[i]})

# economic allocation factor of C6A
C6A_AF = ws_mass['B63'].value

# read power profile and price
ws_windprofile = wb1["Electricity profile & pricing"]
windpower_column = ws_windprofile['E']
windpower_list = [round(windpower_column[x].value, 2) for x in range(2, 8762)] # wind power profile
windpowerprice_column = ws_windprofile['G']
windpower_price = [windpowerprice_column[x].value for x in range(2, 8762)]
hybridpower_column = ws_windprofile['O']
hybridpower_list = [round(hybridpower_column[x].value, 2) for x in range(2, 8762)] # hybrid power profile
hybridpowerprice_column = ws_windprofile['Q']
hybridpower_price = [hybridpowerprice_column[x].value for x in range(2, 8762)]

#########################################################
# create empty lists and dictionaries for storing results
shutdown_hours = []
storage_capacity = []  # GJ
storageLi_capex = []  # M€
LPC = []  # levelised production cost of C6A, €/t
Mt_discounted = []  # annual discounted C6A production quantity, Mt
Mt_annual = [] # annual C6A production quantity, kt
It_ratio = [] # share of CAPEX in LPC
OMt_ratio = [] # share of operation and maintenance costs in LPC
Ft_ratio = [] # share of feedstock costs in LPC
Wt_ratio = [] # share of waste treatment costs in LPC
Et_ratio = [] # share of electricity costs in LPC
Ht_ratio = [] # share of utility costs in LPC
totale_consumption = [] # totoal electricity consumption/nominal consumption, %
surpluse_consumption=[] # surplus electricity consumption/nominal consumption, %
lower_LRR = [] # lower bounry of load ratio range, %
upper_LRR = [] # upper boundary of load ratio range, %

############################
# select electricity profile
power_list = windpower_list  # select between hybridpower_list and windpower_list
load_100 = 1  # nominal load ratio
offpeak_wind = 1.17 # available electricity/nominal electricity consumption, %
offpeak_hybrid = 1.38 # available electricity/nominal electricity consumption, %
power_list = hybridpower_list  # select between hybridpower_list and windpower_list
if power_list == hybridpower_list:
    offpeak_power = offpeak_hybrid
else:
    offpeak_power = offpeak_wind
load_100 = 1  # nominal load ratio

##########################
# start operating scheme 2
# shcematic illustration can be found in Appendix A

for sc in range(0, 51, 1): # range of BESS capacity, GJ/h
    minload = min_operatingload  # %
    length = len(power_list)
    i = -1
    operating_list = [] # C6A production quantity/nominal production quantity, %
    shortage_count = 0
    shortage_timeslot = []
    charging_list = [] # battery's charging or discharging profile, GJ/h
    battery_status = []
    totalconsumption = [] # totoal electricity consumption/nominal consumption, %
    
    # for cash flow
    Wt_ML = []
    Wt_ML.append(0)
    Ft_ML = []
    Ft_ML.append(0)
    Et_ML = []
    Et_ML.append(0)
    Mt_ML = []
    Mt_ML.append(0)
    Ht_ML = []
    Ht_ML.append(0)

    Et_ML_annual = 0
    Ft_ML_annual = 0
    Wt_ML_annual = 0
    Mt_ML_annual = 0
    Ht_ML_annual = 0
    Rt_ML_annual = 0 # revenue, for tax calculation

    It_ML_ref[0] = ws_economics['G18'].value  # M€
    It_ML_ref[15] = (ws_ecoassumption['M59'].value + ws_ecoassumption[
        'N59'].value) / 1e6  # replacement of the membrane contactor
    It_ML_ref[30] = - It_ML_ref[0] * salvage_rate

    x = sc
    for i in range(0, 8760, 1):
        if power_list[i] < minload:
            if x >= (minload - power_list[i]) * MLload_100:
                if x + power_list[i] * MLload_100 >= load_100 * MLload_100:
                    operating_list.append(load_100)
                    charging_list.append(-(load_100 - power_list[i]) * MLload_100)
                    totalconsumption.append(power_list[i])
                    x = x - (load_100 - power_list[i]) * MLload_100
                else:
                    operating_list.append(power_list[i] + round(x / MLload_100, 2))
                    charging_list.append(-x)
                    totalconsumption.append(power_list[i])
                    x = 0
            else:
                operating_list.append(0)
                shortage_count += 1
                shortage_timeslot.append(i)
                charging_list.append(0)
                totalconsumption.append(0)
                x = x
        elif minload <= power_list[i] < load_100:
            if x + power_list[i] * MLload_100 >= load_100 * MLload_100:
                operating_list.append(load_100)
                charging_list.append(-(load_100 - power_list[i]) * MLload_100)
                totalconsumption.append(power_list[i])
                x = x + (power_list[i] - load_100) * MLload_100
            else:
                operating_list.append(power_list[i] + round(x / MLload_100, 2))
                charging_list.append(-x)
                totalconsumption.append(power_list[i])
                x = 0
        else:
            operating_list.append(load_100)
            if x >= sc:
                charging_list.append(0)
                totalconsumption.append(load_100)
                x = x
            else:
                if (power_list[i] - load_100) * MLload_100 + x >= sc:
                    if power_list[i] >= offpeak_power:
                        charging_list.append(sc - x)
                        totalconsumption.append(load_100 + round((sc - x) / MLload_100, 2))
                        x = sc
                    else:
                        charging_list.append(0)
                        totalconsumption.append(load_100)
                        x = x
                else:
                    if power_list[i] >= offpeak_power:
                        charging_list.append((power_list[i] - load_100) * MLload_100)
                        totalconsumption.append(power_list[i])
                        x = x + (power_list[i] - load_100) * MLload_100
                    else:
                        charging_list.append(0)
                        totalconsumption.append(load_100)
                        x = x
        battery_status.append(x)

    ###################################
    # calculate the electricity consumption
    totale_consumption.append(sum(t for t in totalconsumption)/sum(t for t in power_list))
    v = 0
    ec = 0
    for i in range(0, 8760):
        if power_list[i] > load_100:
            v = v + (power_list[i] - load_100)
            ec = ec + (totalconsumption[i]-load_100)
    surpluse_consumption.append(ec/v)

    ############################
    # find the boundaries of LRR
    for n in arange(0.7, round(min(power_list)-0.01, 2), -0.01):
        b = 0
        c = 0
        for i in range(0, 8760):
            if round(power_list[i], 2) == round(n, 2):
                b += 1
                if round(operating_list[i], 2) > 0:
                    c += 1
        if b == c:
            continue
        else:
            lower_LRR.append(round(n + 0.01, 2))
            break
    for n in arange(round(load_100, 2), round(max(power_list), 2), 0.01):
        b = 0
        c = 0
        for i in range(0, 8760):
            if round(power_list[i], 2) == round(n, 2):
                b += 1
                if round(totalconsumption[i], 2) == round(n, 2):
                    c += 1
        if b == c:
            continue
        else:
            upper_LRR.append(round(n-0.01, 2))
            break

    #####################
    # econoimc calcultion
    for n in range(0, 8760):
        if (operating_list[n] != 1) and (operating_list[n] != 0):
            operating_list[n] = round(operating_list[n], 2)
        if power_list == windpower_list:
            Et_ML_annual += totalconsumption[n] * windpower_price[n] * MLload_100 / 1e6  # unit M€
        else:
            Et_ML_annual += totalconsumption[n] * hybridpower_price[n] * MLload_100 / 1e6
        Ft_ML_annual += totalconsumption[n] * Ft_ML_load100 / 1e6  # unit M€
        Wt_ML_annual += operating_list[n] * Wt_ML_load100 / 1e6  # M€
        Mt_ML_annual += operating_list[n] * Mt_ML_load100 / 1e9  # unit Mt
        Ht_ML_annual += (operating_list[n]) * (Ht_list1[operating_list[n]] + Ht_list2[operating_list[n]]) / 1e6  # M€
        Rt_ML_annual += operating_list[n] * ws_mass['B58'].value / 1e6  # M€

    for t in range(lifetime):
        Wt_ML.append(Wt_ML_annual)
        Ft_ML.append(Ft_ML_annual)
        Et_ML.append(Et_ML_annual)
        Mt_ML.append(Mt_ML_annual)
        Ht_ML.append(Ht_ML_annual)

    storageLi_cost = sc * storage_info["unit_cost"][0] / storage_info["efficiency"][0]
    storageLi_capex.append(round(storageLi_cost, 2))
    It_ML_ref[15] += storageLi_cost * (1 + 1/4.31*0.75) - storageLi_cost * storage_info["salvage_rate"][0]
    It_ML_ref[0] += storageLi_cost * (1 + 1/4.31*0.75)
    It_ML_ref[30] += - storageLi_cost * storage_info["salvage_rate"][0]
    OMt = OMt_ML_load100 + (1 + 0.2 + 0.6) * 0.06 * storageLi_cost  # extra OMt due to batteries, extra labour neglected
    if 0 < Rt_ML_annual - (Ft_ML_annual + Ht_ML_annual + Wt_ML_annual + OMt + Et_ML_annual) <= tax_threshold:
        OMt += (Rt_ML_annual - (Ft_ML_annual + Ht_ML_annual + Wt_ML_annual + OMt + Et_ML_annual)) * low_tax_rate
    elif Rt_ML_annual - (Ft_ML_annual + Ht_ML_annual + Wt_ML_annual + OMt + Et_ML_annual) > tax_threshold:
        OMt += tax_threshold * low_tax_rate + (Rt_ML_annual - tax_threshold) * high_tax_rate
    OMt_ML_ref = [OMt for i in range(lifetime + 1)]
    OMt_ML_ref[0] = 0

    LPC_Mt_C6A_ML = npf.npv(discount_rate, Mt_ML)
    LPC_It_ML = C6A_AF * npf.npv(discount_rate, It_ML_ref) / LPC_Mt_C6A_ML
    LPC_Wt_ML = C6A_AF * npf.npv(discount_rate, Wt_ML) / LPC_Mt_C6A_ML
    LPC_Ft_ML = C6A_AF * npf.npv(discount_rate, Ft_ML) / LPC_Mt_C6A_ML
    LPC_Ht_ML = C6A_AF * npf.npv(discount_rate, Ht_ML) / LPC_Mt_C6A_ML
    LPC_Et_ML = C6A_AF * npf.npv(discount_rate, Et_ML) / LPC_Mt_C6A_ML
    LPC_OMt_ML = C6A_AF * npf.npv(discount_rate, OMt_ML_ref) / LPC_Mt_C6A_ML
    LPC_ML = LPC_It_ML + LPC_Wt_ML + LPC_Ft_ML + LPC_Ht_ML + LPC_Et_ML + LPC_OMt_ML

    LPC.append(round(LPC_ML, 2))
    Mt_discounted.append(round(LPC_Mt_C6A_ML* 1e3, 2))
    Mt_annual.append(Mt_ML_annual * 1e3)
    It_ratio.append(LPC_It_ML / LPC_ML)
    OMt_ratio.append(LPC_OMt_ML / LPC_ML)
    Ft_ratio.append(LPC_Ft_ML / LPC_ML)
    Wt_ratio.append(LPC_Wt_ML / LPC_ML)
    Et_ratio.append(LPC_Et_ML / LPC_ML)
    Ht_ratio.append(LPC_Ht_ML / LPC_ML)

    ###########################
    # count the shutdown hours
    shutdown_hours.append(shortage_count)

    ###########################
    # record the storage capacity
    storage_capacity.append(sc)

##########################################################################
# all the information is generated
# assemble the data needed into a dataframe and write to a new excel sheet
# make plots as needed
##########################################################################


