# explorative modelling: Cases D and E
import pandas as pd
import numpy as np
import numpy.random
from numpy import arange
import numpy_financial as npf
from numpy_financial import npv, irr
import xlwt
from openpyxl import load_workbook
import os
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
import sys
import math

# locate file and load the workbook
directory = "/home/jessieluo" # for example
os.chdir(directory)
file_name = "process results.xlsm"
file = directory + "/" + file_name
wb1 = load_workbook(file_name, keep_vba=True, data_only=True)
wb = load_workbook(file, keep_vba=True)

# plant information
min_operatingload = 0.70  # based on the volume flexibility results at the equipment level
ws_elec = wb1['Overall electricity balance']
MLload_100 = ws_elec['A27'].value  # nominal electricity consumption, GJ/h
ws_economics = wb1["CAPEX & OPEX"]
ws_ecoassumption = wb1['Economic assumptions'] # a hidden spreadsheet
lifetime = ws_ecoassumption['B7'].value  # life time, y
discount_rate = ws_ecoassumption['B9'].value
salvage_rate = ws_ecoassumption['B10'].value
projection_2030 = 1  # from NREL
storage_info = {"type": ["Li", "Lead", "RFB"],
                "unit_cost": [1200/3.6/1000*projection_2030, 500/3.6/1000*projection_2030, 1000/3.6/1000*projection_2030],
                "efficiency": [0.9, 0.85, 0.85],
                "salvage_rate": [0.05, 0.05, 0.05]} # unit cost: M€/GJ

# read CAPEX and OPEX parameters
# It and OMt are fixed at the 100% load
It_ML_ref = [0 for i in range(lifetime+1)] # CAPEX flow, M€/y
It_ML_ref[0] = ws_economics['G18'].value  # specify the initial CAPEX, M€/y
It_ML_ref[15] = (ws_ecoassumption['M59'].value + ws_ecoassumption['N59'].value) / 1e6  # specify the replacement of the membrane contactor
It_ML_ref[30] = - It_ML_ref[0] * salvage_rate  # specify the salvage value M€
OMt_ML_load100 = ws_economics['G22'].value # operation and maintenance cost at 100% load, M€/y

# Wt, Ft, Et are always linear to production rate; no need to be adjusted based on power profile
ws_mass = wb1['Overall mass balance']
Wt_ML_load100 = ws_mass['B57'].value  # waste treatment cost at 100% load, €/h
Ft_ML_load100 = ws_mass['B56'].value  # feedstock cost at 100% load, €/h
Et_ML_load100 = ws_elec['A31'].value  # electricity cost at 100% load, €/h
Mt_ML_load100 = ws_mass['B65'].value  # production of C6A at 100% load, kg/h

# Heat consumption is nonlinear to production rate, €/h
ws_heat = wb1["Overall utility balance"]
LR_list = []
for row in ws_heat['B72':'AF72']:
    for x in row:
        LR_list.append(x.value)
Ht_cost_list1 = []
for row in ws_heat['B81':'AF81']:
    for x in row:
        Ht_cost_list.append(round(x.value, 2))
Ht_list = {0: 0}
for i in range(len(LR_list)):
    Ht_list.update({LR_list[i]: Ht_cost_list[i]})

# economic allocation factor of C6A
C6A_AF = ws_mass['B63'].value

# read power profile and price
ws_windprofile = wb1["Electricity profile & pricing"]
hybridpower_column = ws_windprofile['O']
hybridpower_list = [round(hybridpower_column[x].value, 2) for x in range(2, 8762)] # hybrid power profile
hybridpowerprice_column = ws_windprofile['Q']
hybridpower_price = [hybridpowerprice_column[x].value for x in range(2, 8762)]


# create empty lists and dictionaries for storing results
shutdown_hours = []
ST_capacity = []  # L
ST_capex = [] # k€
LPC = []  # levelised production cost of C6A, €/t
Mt_discounted = []  # annual discounted C6A production quantity, Mt
Mt_annual = [] # annual C6A production quantity, kt
It_ratio = [] # share of CAPEX in LPC
OMt_ratio = [] # share of operation and maintenance costs in LPC
Ft_ratio = [] # share of feedstock costs in LPC
Wt_ratio = [] # share of waste treatment costs in LPC
Et_ratio = [] # share of electricity costs in LPC
Ht_ratio = [] # share of utility costs in LPC
totale_consumption = [] # totoal electricity consumption/nominal consumption, %
lower_LRR = [] # lower bounry of load ratio range, %
upper_LRR = [] # upper boundary of load ratio range, %

############################
# select electricity profile
power_list = hybridpower_list
load_100 = 1  # nominal load ratio
convert = 64066.79  # 64066.79 L/h

##########################
# start
# shcematic illustration can be found in Appendix A
for st in arange(0, 50002, 1000):
    minload = min_operatingload  # capacity factor
    length = len(power_list)
    i = -1
    operating_list = []
    shortage_count = 0
    shortage_timeslot = []
    storing_list = []
    ST_status = []
    totalconsumption = []


    Wt_ML = []
    Wt_ML.append(0)
    Ft_ML = []
    Ft_ML.append(0)
    Et_ML = []
    Et_ML.append(0)
    Mt_ML = []
    Mt_ML.append(0)
    Ht_ML = []
    Ht_ML.append(0)

    Et_ML_annual = 0
    Ft_ML_annual = 0
    Wt_ML_annual = 0
    Mt_ML_annual = 0
    Ht_ML_annual = 0
    Rt_ML_annual = 0 # revenue, for tax calculation


    It_ML_ref[0] = ws_economics['G18'].value  # M€
    It_ML_ref[15] = (ws_ecoassumption['M59'].value + ws_ecoassumption[
        'N59'].value) / 1e6  # replacement of the membrane contactor
    It_ML_ref[30] = - It_ML_ref[0] * salvage_rate

    s = st
    while i < length - 1:
        i += 1
        if power_list[i] < minload:  # not able to operate by power
            if round((minload - power_list[i]) * convert, 2) <= s:  # st can support
                operating_list.append(minload)
                totalconsumption.append(power_list[i])
                storing_list.append(-round((minload - power_list[i]) * convert, 2))
                s = s - round((minload - power_list[i]) * convert, 2)
            else:  # st cannot support
                operating_list.append(0)
                shortage_count += 1
                if st - s > 0:
                    if round(power_list[i] * convert, 2) > st - s:
                        storing_list.append(st - s)
                        totalconsumption.append(round((st-s) / convert, 2))
                        s = st
                    else:
                        storing_list.append(round(power_list[i] * convert, 2))
                        totalconsumption.append(power_list[i])
                        s = s + round(power_list[i] * convert, 2)
                else:
                    storing_list.append(0)
                    totalconsumption.append(0)
                    s = s
        elif minload <= power_list[i] < load_100:
            if st - s > 0:
                if round((power_list[i] - minload) * convert, 2) > st - s:
                    storing_list.append(st - s)
                    operating_list.append(power_list[i] - round((st - s)/convert, 2))
                    totalconsumption.append(power_list[i])
                    s = st
                else:
                    operating_list.append(minload)
                    totalconsumption.append(power_list[i])
                    storing_list.append(round((power_list[i] - minload) * convert, 2))
                    s = s + round((power_list[i] - minload) * convert, 2)
            else:
                operating_list.append(power_list[i])
                totalconsumption.append(power_list[i])
                storing_list.append(0)
                s = s
        else:  # power_list[i] >= load_100:
            totalconsumption.append(load_100)
            if st - s > 0:
                if st - s > round((power_list[i] - minload) * convert, 2):
                    operating_list.append(minload)
                    storing_list.append(round((load_100 - minload) * convert, 2))
                    s = s + round((load_100 - minload) * convert, 2)
                else:
                    if round((load_100 - minload) * convert, 2) >= st - s:
                        operating_list.append(load_100 - round((st - s)/convert, 2))
                        storing_list.append(st - s)
                        s = st
                    else:
                        operating_list.append(minload)
                        storing_list.append(round((load_100 - minload) * convert, 2))
                        s = s + round((load_100 - minload) * convert, 2)

            else:
                operating_list.append(load_100)
                storing_list.append(0)
                s = s

        ST_status.append(round(s, 2))

    #######################################
    # calculate the electricity consumption
    totale_consumption.append(sum(t for t in totalconsumption) / sum(t for t in power_list))

    ############################
    # find the boundaries of LRR
    for n in arange(0.7, round(min(power_list)-0.01, 2), -0.01):
        b = 0
        c = 0
        for i in range(0, 8760):
            if round(power_list[i], 2) == round(n, 2):
                b += 1
                if round(operating_list[i], 2) > 0:
                    c += 1
        if b == c:
            continue
        else:
            lower_LR_at.append(round(n + 0.01, 2))
            break
    for n in arange(round(load_100, 2), round(max(power_list), 2), 0.01):
        b = 0
        c = 0
        for i in range(0, 8760):
            if round(power_list[i], 2) == round(n, 2):
                b += 1
                if round(totalconsumption[i], 2) == round(n, 2):
                    c += 1
        if b == c:
            continue
        else:
            upper_LR_at.append(round(n-0.01, 2))
            break

    #####################
    # econoimc calcultion
    for n in range(0, 8760):
        if (operating_list[n] != 1) and (operating_list[n] != 0):
            operating_list[n] = round(operating_list[n], 2)
        if power_list == windpower_list:
            Et_ML_annual += totalconsumption[n] * windpower_price[n] * MLload_100 / 1e6  # unit M€
        else:
            Et_ML_annual += totalconsumption[n] * hybridpower_price[n] * MLload_100 / 1e6
        Ft_ML_annual += totalconsumption[n] * Ft_ML_load100 / 1e6  # unit M€
        Wt_ML_annual += operating_list[n] * Wt_ML_load100 / 1e6  # M€
        Mt_ML_annual += operating_list[n] * Mt_ML_load100 / 1e9  # unit Mt
        Ht_ML_annual += (operating_list[n]) * (Ht_list1[operating_list[n]] + Ht_list2[operating_list[n]]) / 1e6  # M€
        Rt_ML_annual += operating_list[n] * ws_mass['B58'].value / 1e6  # M€

    for t in range(lifetime):
        Wt_ML.append(Wt_ML_annual)
        Ft_ML.append(Ft_ML_annual)
        Et_ML.append(Et_ML_annual)
        Mt_ML.append(Mt_ML_annual)
        Ht_ML.append(Ht_ML_annual)

    if st < 10000:
        ST_cost = (4.31) * 10 * st * 2 / 1e6
    else:
        ST_cost = (4.31) * 10 * (5800 + 1600 * (st / 1000) ** 0.7) / 1e6
    ST_capex.append(round(ST_cost*1e3, 2))
    It_ML_ref[0] += ST_cost * (1 + 1/4.31*0.75)
    It_ML_ref[15] += ST_cost * (1 + 1/4.31*0.75) - ST_cost * salvage_rate
    It_ML_ref[30] += - ST_cost * salvage_rate
    OMt = OMt_ML_load100 + (1 + 0.2 + 0.6) * 0.06 * ST_cost  # extra OMt due to batteries, extra labour neglected
    if 0 < Rt_ML_annual - (Ft_ML_annual + Ht_ML_annual + Wt_ML_annual + OMt + Et_ML_annual) <= tax_threshold:
        OMt += (Rt_ML_annual - (Ft_ML_annual + Ht_ML_annual + Wt_ML_annual + OMt + Et_ML_annual)) * low_tax_rate
        print("yes low")
    elif Rt_ML_annual - (Ft_ML_annual + Ht_ML_annual + Wt_ML_annual + OMt + Et_ML_annual) > tax_threshold:
        OMt += tax_threshold * low_tax_rate + (Rt_ML_annual - tax_threshold) * high_tax_rate
        print("yes high")
    else:
        OMt = OMt
        print("no tax")
    OMt_ML_ref = [OMt for i in range(lifetime + 1)]
    OMt_ML_ref[0] = 0

    LPC_Mt_C6A_ML = npf.npv(discount_rate, Mt_ML)
    LPC_It_ML = C6A_AF * npf.npv(discount_rate, It_ML_ref) / LPC_Mt_C6A_ML
    LPC_Wt_ML = C6A_AF * npf.npv(discount_rate, Wt_ML) / LPC_Mt_C6A_ML
    LPC_Ft_ML = C6A_AF * npf.npv(discount_rate, Ft_ML) / LPC_Mt_C6A_ML
    LPC_Ht_ML = C6A_AF * npf.npv(discount_rate, Ht_ML) / LPC_Mt_C6A_ML
    LPC_Et_ML = C6A_AF * npf.npv(discount_rate, Et_ML) / LPC_Mt_C6A_ML
    LPC_OMt_ML = C6A_AF * npf.npv(discount_rate, OMt_ML_ref) / LPC_Mt_C6A_ML
    LPC_ML = LPC_It_ML + LPC_Wt_ML + LPC_Ft_ML + LPC_Ht_ML + LPC_Et_ML + LPC_OMt_ML

    LPC.append(round(LPC_ML, 2))
    Mt_discounted.append(round(LPC_Mt_C6A_ML * 1e3, 2))
    Mt_annual.append(Mt_ML_annual*1e3)
    It_ratio.append(LPC_It_ML / LPC_ML)
    OMt_ratio.append(LPC_OMt_ML / LPC_ML)
    Ft_ratio.append(LPC_Ft_ML / LPC_ML)
    Wt_ratio.append(LPC_Wt_ML / LPC_ML)
    Et_ratio.append(LPC_Et_ML / LPC_ML)
    Ht_ratio.append(LPC_Ht_ML / LPC_ML)

    ###########################
    # count the shutdown hours
    shutdown_hours.append(shortage_count)

    ###########################
    # record the storage capacity
    storage_capacity.append(sc)

##########################################################################
# all the information for case D is generated
# assemble the data needed into a dataframe and write to a new excel sheet
# make plots as needed
##########################################################################


#################################################################################
# error check for assumptions made
# calculate how much production would be reduced if the assumptions were not made
# !!! make sure the codes aboven only coorespond to the D-ST35 !!!

s = 35000 # !!!!!!!!!!!!!!!
less_production = 0
ST_status.insert(0, s)
ST_status.pop()

for i,j,z,h in zip(storing_list, power_list,operating_list, ST_status):
    if -i > 0 and z >= minload and (j*MLload_100 - (h/convert * 1.12))/MLload_100 + h/convert >= minload:
        less_production = less_production + (-i/convert*1.13/67.64)*1.155
    elif -i > 0 and z >= minload and (j*MLload_100 - (h/convert * 1.12))/MLload_100 + h/convert < minload:
        less_production = less_production + 0.7 * 1.155



##############################################################
# for Case E
# calculate extra production required for paying back the BESS
# !!! make sure the codes aboven only coorespond to the D-ST43 !!!
s = 43000 # !!!!!!!!!!!!!!!
Mt_extra = []
def myfunction(BESS, mode="default"):
    I_BESS = [0 for i in range(lifetime + 1)]
    I_BESS[0] = 1e6 * BESS * storage_info["unit_cost"][0] / storage_info["efficiency"][0] * (1 + 1 / 4.31 * 0.75)

    I_BESS[15] = 1e6 * BESS * storage_info["unit_cost"][0] / storage_info["efficiency"][0] * \
                 (1 + 1 / 4.31 * 0.75 - storage_info["salvage_rate"][0])
    I_BESS[30] = - 1e6 * BESS * storage_info["unit_cost"][0] / storage_info["efficiency"][0] * \
                 storage_info["salvage_rate"][0]
    OM_BESS = (1 + 0.2 + 0.6) * 0.06 * \
              (1e6 * BESS * storage_info["unit_cost"][0] / storage_info["efficiency"][0])
    EAC = 1e6 * BESS * storage_info["unit_cost"][0] / storage_info["efficiency"][0] * (
                1 + 1 / 4.31 * 0.75) * discount_rate \
          / (1 - (1 + discount_rate) ** (-lifetime / 2-1))
    if mode == "default":
        return (EAC + OM_BESS) * 0.947 / (
                    LPC[0] - (Ft_ML_load100 + Wt_ML_load100 + Ht_cost_list1[0] + Ht_cost_list2[0] + Et_ML_load100) * 0.947/ (
                        Mt_ML_load100 / 1e3))
    else:
        return EAC

potential_BESS = []
sc_range = arange(0, 21, 1)
for BESS in sc_range:
    potential1 = 0
    potential2 = 0
    potential3 = 0
    Mt_extra.append(myfunction(BESS, mode="default"))
    for i in range(0, 8670, 1):
        if minload - round(BESS / MLload_100, 2) <= power_list[i] < minload:
            potential1 += power_list[i] + round(BESS / MLload_100, 2)
        elif minload <= power_list[i] <= load_100 - round(BESS / MLload_100, 2):
            potential2 += round(BESS / MLload_100, 2)
        elif load_100 - round(BESS / MLload_100, 2) < power_list[i] <= load_100:
            potential3 += load_100 - power_list[i]
    potential_BESS.append(round(potential1 + potential2 + potential3, 2))

#####################################
# extra production possible with BESS
potential_production_list = [] # extra production possible with BESS, t/y
potential_energy_list = [] # extra electricity taken in possible with BESS, %
for sc in range(1, 21, 1):
    x = sc
    potential1 = 0
    potential2 = 0
    for i in range(0, 8760, 1):
        if power_list[i] > load_100:
            if sc > x:
                if sc - x <= round((power_list[i] - load_100) * MLload_100, 2):
                    x = sc
                else:
                    x = x + round((power_list[i] - load_100) * MLload_100, 2)
        else:
            if minload <= operating_list[i] < load_100 and x > 0:
                if round((load_100 - operating_list[i]) * MLload_100, 2) >= x:
                    potential1 += round(x / MLload_100, 2)
                    x = 0
                else:
                    x = x - round((load_100 - operating_list[i]) * MLload_100, 2)
                    potential1 += load_100 - operating_list[i]

            elif power_list[i] < minload and power_list[i] + round(x / MLload_100, 2) >= minload:
                potential2 += power_list[i] + round(x / MLload_100, 2)
                x = 0

    potential_production = (potential1 + potential2 + round(31 / MLload_100, 2)) * Mt_ML_load100 / 1e3
    potential_production_list.append(potential_production)

# data used for making the plots are Mt_extra and potential_production_list

